package meerkat.mixer.main;

import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.crypto.mixnet.Mixer;
import meerkat.crypto.mixnet.MixerOutput;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import meerkat.mixer.mixing.MixNetwork;
import meerkat.mixer.necessary.AsyncBulletinBoardClient;


import java.util.ArrayList;
import java.util.List;
import java.util.Random;


/**
 * Created by Tzlil on 12/17/2015.
 * this class define all the operation meerkat.mixer.mixing party should do:
 *  1. receive previous mixers output (re encrypted votes + proofs)
 *  2. verify its input
 *  3. mix
 *  4. send the meerkat.mixer.mixing output
 */
public class MainMixing {

    private final Mixer mixer;
    private final Mix2ZeroKnowledgeVerifier verifier;
    private final int n, layers;
    private final AsyncBulletinBoardClient asyncBulletinBoardClient;
    private final byte[] id;


    /**
     * constructor
     * @param mixer
     * @param verifier
     * @param n
     * @param asyncBulletinBoardClient
     * @param id
     */
    public MainMixing(Mixer mixer, Mix2ZeroKnowledgeVerifier verifier, int n
            , AsyncBulletinBoardClient asyncBulletinBoardClient, byte[] id) {
        this.mixer = mixer;
        this.verifier = verifier;
        this.n = n;
        this.layers = MixNetwork.numberOfLayers(n);
        this.asyncBulletinBoardClient = asyncBulletinBoardClient;
        this.id = id;
    }

    /**
     *
     * @param prevBatchIds
     * @param batchId
     * @param random
     * @param callback
     * @throws Throwable
     */
    public void main(List<Integer> prevBatchIds, int batchId, Random random, AsyncBulletinBoardClient.ClientCallback<?> callback) throws Throwable {

        List<Crypto.RerandomizableEncryptedMessage> mixerInput;

        List<BatchHandler> batchHandlers = new ArrayList<BatchHandler>(prevBatchIds.size());
        BatchHandler currentBatchHandler;
        for (Integer prevBatchId : prevBatchIds) {
            currentBatchHandler = new BatchHandler(n, layers,verifier);
            asyncBulletinBoardClient.readBatch(id, prevBatchId,currentBatchHandler);
            batchHandlers.add(currentBatchHandler);
        }
        // check all handlers messages were received
        for (BatchHandler batchHandler : batchHandlers) {
            synchronized (batchHandler){
                if(!batchHandler.isMsgReceived())
                    batchHandler.wait();
            }
        }
        // assert all handlers succeeded
        for (BatchHandler batchHandler : batchHandlers) {
            if(!batchHandler.verifyTable()){
                throw new Exception("invalid input");
            }
        }

        BatchHandler lastBatchHandler = batchHandlers.get(batchHandlers.size() - 1);
        mixerInput = lastBatchHandler.getInputForMixer();

        MixerOutput mixerOutput = mixer.mix(mixerInput,random);
        updateBB(mixerOutput, batchId, callback);

    }

    /**
     * send meerkat.mixer.mixing output to BB
     * @param mixerOutput
     * @param batchId
     * @param callback
     */
    private void updateBB(MixerOutput mixerOutput
            , int batchId, AsyncBulletinBoardClient.ClientCallback<?> callback) {

        BatchConverter batchConverter = new BatchConverter(n,layers);
        List<BulletinBoardAPI.BatchChunk> batchChunkList = batchConverter.MixerOutput2BatchChunk(mixerOutput);
        asyncBulletinBoardClient.postBatch(id, batchId, batchChunkList, callback);
    }

}
